package intensity;
import java.io.BufferedWriter;
import java.io.File;
import java.io.IOException;
import java.text.DecimalFormat;
import java.util.Iterator;

import org.apache.commons.math3.util.Pair;

import weka.classifiers.evaluation.NominalPrediction;
import weka.classifiers.evaluation.ThresholdCurve;
import weka.core.FastVector;
import weka.core.Instances;

/**
 * Classification Evaluation
 */
@SuppressWarnings("deprecation")
public class ClassificationEvaluation {	
	/**
	 * member variables required for keeping track
	 * of statistics
	 */
	private int nInstances = 0;
	private int nErrors = 0;
	
	/**
	 * confusion matrix
	 * first index: actual classification
	 * second index: predicted classification
	 */
	private int[][] confusionMatrix;
	
	/**
	 * prediction scores
	 */
	FastVector<NominalPrediction> predictionScores;
	
	private static DecimalFormat numberFormatter = new DecimalFormat("#0.0000");
	
	/**
     * Constructor
     * @throws IOException 
     */
    public ClassificationEvaluation() 
        throws IOException
    {
        this(null, 2);
    }
	
	/**
	 * Constructor
	 * @param singleResultLog file to log individual correctness results to
	 * @param nClasses  number of classes
	 * @throws IOException 
	 */
	public ClassificationEvaluation(File singleResultLog, int nClasses) 
        throws IOException
    {
		confusionMatrix = new int[nClasses][nClasses];
		predictionScores = new FastVector<NominalPrediction>();
	}
	
	/**
	 * update evaluation by taking a new instance into account
	 * @param trueClassification the true classification
	 * @param predictedClassification the classification predicted by the machine learning
	 *   classifier
	 * @throws Exception
	 */
	public final void updateEvaluation(int trueClassification, int predictedClassification) throws Exception {
		nInstances++;
		
		boolean isCorrect = (trueClassification == predictedClassification);
		if (!isCorrect)
			nErrors++;
		confusionMatrix[trueClassification][predictedClassification]++;
	}
	
	/**
	 * update evaluation by taking a new instance into account
	 * @param trueClassification the true classification
	 * @param predictedClassification the classification predicted by the machine learning
	 *   classifier
	 * @param predictedClassificationDist  classification distribution predicted
	 * @throws Exception
	 */
	public final void updateEvaluation(int trueClassification, int predictedClassification, double[] predictedClassificationDist) throws Exception {
		updateEvaluation(trueClassification, predictedClassification);
		predictionScores.addElement(
				new NominalPrediction(trueClassification, predictedClassificationDist, 1.0));
	}
	
	/**
	 * @return number of true positives (for 2-class classifications only)
	 */
	public final int truePositives() {
		return confusionMatrix[1][1];
	}
	
	/**
	 * @return number of false positives (for 2-class classifications only)
	 */
	public final int falsePositives() {
		return confusionMatrix[0][1];
	}
	
	/**
	 * @return number of false negatives (for 2-class classifications only)
	 */
	public final int falseNegatives() {
		return confusionMatrix[1][0];
	}
	
	/**
	 * @return number of true negatives (for 2-class classifications only)
	 */
	public final int trueNegatives() {
		return confusionMatrix[0][0];
	}
	
    public int getNAgreements() {
        int nAgreements = 0; 
        for (int i = 0; i < confusionMatrix.length; i++)
            nAgreements += confusionMatrix[i][i];
        return nAgreements;
    }
    
    public int getNTotal() {
        int nItems = 0; 
        for (int i = 0; i < confusionMatrix.length; i++)
            for (int j = 0; j < confusionMatrix.length; j++)
                nItems += confusionMatrix[i][j];
        return nItems;
    }
	
	/**
	 * @return precision (i.e. which percentage of all positively predicted
	 *  instances are actually positive instances)  (for 2-class classifications only)
	 */
	public final double precision() {
		// check if total number of documents predicted as negative is zero
		if (confusionMatrix[1][1] + confusionMatrix[0][1] == 0) {
			// if there are no documents that are positively
			// then return 1 
			if (confusionMatrix[1][1] == 0)
				return 1;
			else
				return 0;
		}
		return ((double) confusionMatrix[1][1]) / (confusionMatrix[1][1] + confusionMatrix[0][1]);
	}
	
	/**
	 * @return recall (i.e. which percentage of all actually positive instances were
	 *   predicted to be positive instances)   (for 2-class classifications only)
	 */
	public final double recall() {
		// if there are no positive examples then return 1
		if (confusionMatrix[1][1] + confusionMatrix[1][0] == 0)
			return 1;
		return ((double) confusionMatrix[1][1]) / (confusionMatrix[1][1] + confusionMatrix[1][0]);
	}
	
	/**
	 * @return F1 measure  (for 2-class classifications only)
	 */
	public final double f1() {
		double precision = precision();
		double recall = recall();
		if (precision + recall == 0)
			return 0;
		return (2 * precision * recall) / (precision + recall);
	}
	
	/**
	 * compute f-scores
	 * @param beta beta of f-score formula (non-negative)
	 * @return F-score measure (for 2-class classifications only)
	 */
	public final double fScore(double beta) {
		double precision = precision();
		double recall = recall();
		if (beta * precision == 0 && recall == 0)
			// the denominator will become zero if beta or precision
			// are zero and recall is zero
			return 0;
		return ((1.0 + (beta * beta)) * precision * recall) / (beta * beta * precision + recall);
	}
	
	
	/**
	 * compute f-score
	 * @param beta beta of f-score formula (non-negative)
	 * @param precision precision
	 * @param recall recall
	 * @return f-score
	 */
	public static double fScore(double beta, double precision, double recall) {
		return (1.0 + beta * beta) * (precision * recall) / (beta * beta * precision + recall);
	}
	
	/**
	 * @return error rate
	 */
	public final double errorRate() {
		return ((double) nErrors) / nInstances;
	}
	
	/**
	 * Cohen's Kappa
	 * Works for arbitrary-sized confusion matrices, i.e. two raters classifying objects into an arbitrary set of categories,
	 * as long as these categories are shared, resulting in a standard square confusion matrix. One annotator's choices are
	 * represented by the rows, and another annotator's choices for the same categories are represented by the columns.
	 * Cohen's Kappa does not, however, apply to more than two annotators, i.e. confusion tensors with more than 2 dimensions.
	 * @return score in [0,1]
	 */
	public double getCohenKappa() {
	    // compute observed agreement probability
	    int totalObservations = getNTotal();
	    double pAgreement = (totalObservations > 0) ? ((double) getNAgreements())/totalObservations : 0.0;
	    
	    // get marginals for computing chance agreement: These express how likely an annotator randomly
	    // picks a given category
	    int nCats = confusionMatrix.length; 
	    double[] rowMarginals = new double[nCats];
        double[] colMarginals = new double[nCats];
	    double totalSum = 0.0;
	    for (int i = 0; i < nCats; i++) {
            double rowSum = 0.0;
            for (int j = 0; j < nCats; j++)
                rowSum += confusionMatrix[i][j];
            rowMarginals[i] = rowSum;
            totalSum += rowSum;
	    }
        if (totalSum == 0.0)
            return 0.0;
	    // compute column marginals
	    for (int j = 0; j < nCats; j++) {
            double colSum = 0.0;
            for(int i = 0; i < nCats; i++)
                colSum += confusionMatrix[i][j];
            colMarginals[j] = colSum;
	    }
	    
	    // compute chance agreement probability P(E)
	    double squaredSum = totalSum * totalSum;
        double pE = 0.0;
	    for (int i = 0; i < nCats; i++)
	        pE += (colMarginals[i] * rowMarginals[i]) / squaredSum;
	    
	    // Cohen's Kappa formula
	    return (pAgreement - pE) / (1 - pE);
	}

	
	/**
	 * Print results of Evaluation
	 * @param out  logger
	 */
	public void printResults(Logger out) {
		if (confusionMatrix.length == 2) {
			out.println("TP: " + truePositives());
			out.println("FP: " + falsePositives());
			out.println("TN: " + trueNegatives());
			out.println("FN: " + falseNegatives());
		}
		double errorRate = errorRate();
		Pair<Double,Double> wilson = wilsonScoreInterval(1.0 - errorRate, nInstances);
		out.println("Accuracy:   " + numberFormatter.format(1.0 - errorRate) + " (Wilson: " + numberFormatter.format(wilson.getKey()) + " ± " + numberFormatter.format(wilson.getValue()) + ")");
		wilson = wilsonScoreInterval(errorRate, nInstances);
		out.println("Error Rate: " + numberFormatter.format(errorRate) + " (Wilson: " + numberFormatter.format(wilson.getKey()) + " ± " + numberFormatter.format(wilson.getValue()) + ")");
		if (confusionMatrix.length == 2) {
			double precision = precision();
			wilson = wilsonScoreInterval(precision, (confusionMatrix[1][1] + confusionMatrix[0][1]));
			out.println("Precision:  " + numberFormatter.format(precision) + " (Wilson: " + numberFormatter.format(wilson.getKey()) + " ± " + numberFormatter.format(wilson.getValue()) + ")");
			double recall = recall();
			wilson = wilsonScoreInterval(recall, (confusionMatrix[1][1] + confusionMatrix[1][0]));
			out.println("Recall:     " + numberFormatter.format(recall) + " (Wilson: " + numberFormatter.format(wilson.getKey()) + " ± " + numberFormatter.format(wilson.getValue()) + ")");
			out.println("F1:         " + numberFormatter.format(f1()));
			out.println("F0.5:       " + numberFormatter.format(fScore(0.5)));
			if (predictionScores.size() > 0)
				out.println("ROC-AUC:    " + numberFormatter.format(getROCAUC()));
		}
	}	
	
	/**
	 * compute Spearman's Rank Correlation Coefficient
	 * 
	 * Cf.
	 * http://en.wikipedia.org/wiki/Spearman%27s_rank_correlation_coefficient
	 * 
	 * @param ranking1 ranking 1
	 * @param ranking2 ranking 2
	 * @return rank correlation coefficient in [-1,1]
	 */
	public static double spearmanRankCorrelationCoefficient(
			MapSortedByValue<String, Double> ranking1, MapSortedByValue<String, Double> ranking2)
	{
		double result = 0.0;
		int r1 = 0;
		Iterator<MapSortedByValue.Entry<String,Double>> it1 = ranking1.sortedEntrySetIterator();
		while (it1.hasNext()) {
			MapSortedByValue.Entry<String, Double> ranking1Entry = it1.next();
			String entryName = ranking1Entry.getKey();
			int r2 = 0;
			Iterator<MapSortedByValue.Entry<String,Double>> it2 = ranking2.sortedEntrySetIterator();
			while (it2.hasNext()) {
				MapSortedByValue.Entry<String, Double> ranking2Entry = it2.next();
				if (ranking2Entry.getKey().equals(entryName))
					break;
				r2++;
			}
			int distance = (r1 - r2);
			distance *= distance;
			result += distance;
			r1++;
		}
		int nEntries = ranking2.size();
		return 1 - (6.0 * result)/(nEntries * (nEntries*nEntries - 1));
	}

	/**
	 * close results overview file
	 * @param overviewFileWriter  BufferedWriter instance to close
	 * @throws IOException
	 */
	public static void closeResultsOverviewFile(BufferedWriter overviewFileWriter) throws IOException {
		if (overviewFileWriter != null)
			overviewFileWriter.close();
	}
	
	/**
	 * Wilson test
	 * @param pEstimate estimate of precision/accuracy based on sample data
	 * @param nSamples  number of samples 
	 * @return pair with centre and delta 
	 */
	public final static Pair<Double,Double> wilsonScoreInterval(double pEstimate, double nSamples) {
		/*
		double numeratorPart1 = pEstimate + zAlpha * zAlpha * (0.5d / n); 
		double numeratorPart2 = zAlpha * Math.sqrt((pEstimate * (1-pEstimate) + zAlpha*zAlpha * (0.25d/n))/n);
		double denominator = 1 + zAlpha*zAlpha/n;
		double lowerLimit = (numeratorPart1 - numeratorPart2) / denominator;
		double upperLimit = (numeratorPart1 + numeratorPart2) / denominator;
		return new Pair<Double,Double>(lowerLimit, upperLimit);
		*/
		double z = 1.95996398454005423552; // Z_{(1-\alpha)/2}, i.e. the 1−alpha/2 percentile of a standard normal distribution
										   // cf. http://en.wikipedia.org/wiki/1.96
        double centre = (pEstimate + 1 / 2.0 / nSamples * z * z) / (1 + 1.0 / nSamples * z * z);
        double d = z * Math.sqrt((pEstimate * (1 - pEstimate) + 1 / 4.0 / nSamples * z * z) / nSamples)
                 / (1 + z * z / nSamples);
        return new Pair<Double,Double>(centre, d);
	}
	
	/**
	 * ROC area under curve
	 * 
	 * Important:
	 * - ROC curves are not the same as precision-recall curves!
	 * - ROC (receiver operating characteristic) curves plot sensitivity
	 *   (recall) on the y axis vs. 1-specificity on the x axis, where 
	 *   (1-specificity) is tn/allneg = tn/(fp+tn)
	 * - scores must have been passed to the update method to use this
	 * 
	 * @return ROC area under curve 
	 */
	public final double getROCAUC() {
		ThresholdCurve thresholdCurve = new ThresholdCurve();
		Instances result = thresholdCurve.getCurve(predictionScores);
		return ThresholdCurve.getROCArea(result);
	}
	
    public void printConfusionMatrix(Logger logger) {
        logger.print("\t");
        for (int j = 0; j < confusionMatrix.length; j++)
            logger.print("(" + j + ")\t");
        logger.println();
        for (int i = 0; i < confusionMatrix.length; i++) {
            logger.print("(" + i + "):\t");
            for (int j = 0; j < confusionMatrix.length; j++) {
                logger.print(confusionMatrix[i][j] + "\t");
            }
            logger.println();
        }
    }
}
