package intensity;


import gnu.trove.THashMap;

import java.util.Map;
import java.util.TreeSet;

/**
 * HashMap for integer counter values
 * 
 * @param <KeyType> type of the objects that are to be counted
 */
public class CountMap<KeyType> extends THashMap<KeyType,Integer> implements Map<KeyType,Integer>
{
	/**
	 * for serializability
	 */
	private static final long serialVersionUID = 1142746263959415310L;

	/**
	 * Increment counter
	 * @param key  key
	 * @param byValue  how much to increment the counter 
	 * @return new counter value
	 */
	public final int increment(KeyType key, int byValue) {
		Integer oldVal = get(key);
		int newAmount;
		if (oldVal != null)
			newAmount = oldVal + byValue;
		else
			newAmount = byValue;
		put(key, newAmount);
		return newAmount;
	}

	/**
	 * Increment counter
	 * @param key key
	 * @return new counter value
	 */
	public final int increment(KeyType key) {
		return increment(key, 1);
	}

	/**
	 * increment counters using another CountMap as a source
	 * @param source source CountMap
	 */
	public final void incrementAll(CountMap<KeyType> source) {
		for (Map.Entry<KeyType,Integer> entry: source.entrySet()) {
			increment(entry.getKey(), entry.getValue());
		}
	}
	
	/**
	 * ensure stored value is not smaller than given value 
	 * @param key  key
	 * @param value  given value
	 * @return value after possible updating
	 */
	public final int ensureMinimumValue(KeyType key, int value) {
		Integer oldVal = get(key);
		if (oldVal == null)
			put(key, value);
		else if (oldVal < value)
			put(key, value);
		else
			return oldVal;
		return value;
	}	
	
	/**
     * print sorted by value
     * @param <T> type of key
     * @param countMap
     * @param logger
     */
    public final static <T extends Comparable<T>> void printSortedByKey(CountMap<T> countMap, Logger logger)
    {
        TreeSet<T> keys = new TreeSet<T>(countMap.keySet());
        for (T key: keys)
            logger.println(key.toString() + ": " + countMap.get(key));
    }
}
