package intensity;
import gnu.trove.THashMap;

import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;

/**
 * A mapping between keys and values with additional methods
 * that allow for retrieving entries sorted by value.
 * 
 * Note that it may be more efficient to use a Fibonacci heap
 * in certain cases.
 *  
 * @param <KeyType> type of keys
 * @param <ValType> type of values
 */
public class MapSortedByValue<KeyType,ValType extends Comparable<ValType>> {
	
	private Map<KeyType,ValType> keyAssociations;
	private Map<KeyType,ValType> keyAssociationsUnmodifiable;
	
	private TreeSet<Entry<KeyType,ValType>> sortedValueKeyPairs;
	private Set<Entry<KeyType,ValType>> sortedValueKeyPairsUnmodifiable;
	
	/**
	 * Construct a new Mapping Structure
	 */
	public MapSortedByValue() {
		keyAssociations = new THashMap<KeyType,ValType>();
		keyAssociationsUnmodifiable = Collections.unmodifiableMap(keyAssociations);
		sortedValueKeyPairs = new TreeSet<Entry<KeyType,ValType>>();
		sortedValueKeyPairsUnmodifiable = Collections.unmodifiableSet(sortedValueKeyPairs);
	}
	
	/**
	 * get value associated with key
	 * @param key key
	 * @return value associated with key or null if none
	 */
	public final ValType get(KeyType key) {
		return keyAssociations.get(key); 
	}
	
	/**
	 * add mapping
	 * @param key key
	 * @param value value to associate with key
	 */
	public final void put(KeyType key, ValType value) {
		if (keyAssociations.containsKey(key))
			// first delete old entry in sortedValueKeyPairs
			sortedValueKeyPairs.remove(new Entry<KeyType,ValType>(key, keyAssociations.get(key)));
		keyAssociations.put(key, value);
		sortedValueKeyPairs.add(new Entry<KeyType,ValType>(key,value));
	}
	
	/**
	 * add mapping if the new value is greater than any original one
	 * @param key key
	 * @param value value to associate with key
	 *
	public void putIfGreater(KeyType key, ValType value) {
		ValType oldValue = keyAssociations.get(key);
		boolean greater;
		if (oldValue != null)
			greater = oldValue.compareTo(value) > 0;
		else
			greater = true;
		if (greater) {
			keyAssociations.put(key, value);
			if (oldValue != null)
				sortedValueKeyPairs.remove(new Entry<KeyType,ValType>(key,oldValue));
			sortedValueKeyPairs.add(new Entry<KeyType,ValType>(key,value));
		}
	}*/
	
	/**
	 * check if a mapping exists
	 * @param key key
	 * @return whether a mapping exists
	 */
	public final boolean containsKey(KeyType key) {
		return keyAssociations.containsKey(key);
	}
	
	/**
	 * get normal (unsorted) entry set
	 * 
	 * Note: cannot be used for deletions or other modifications
	 * 
	 * @return set of entries
	 */
	public final Set<Map.Entry<KeyType,ValType>> entrySet() {
        return keyAssociationsUnmodifiable.entrySet();
    }
	
	/**
	 * get sorted entry set
	 * 
	 * Note: cannot be modified directly.
	 * @return sorted mapping entries
	 */
	public final Collection<Entry<KeyType,ValType>> sortedEntrySet() {
		return sortedValueKeyPairsUnmodifiable; 
	}
	
	/**
	 * get sorted entry set iterator
	 * @return sorted mapping entries
	 */
	public final Iterator<Entry<KeyType,ValType>> sortedEntrySetIterator() {
		return sortedValueKeyPairsUnmodifiable.iterator(); 
	}
	
	/**
     * get reverse iterator
     * @return reverse iterator
     */
    public final Iterator<Entry<KeyType,ValType>> sortedReverseEntrySetIterator() {
        return sortedValueKeyPairs.descendingIterator();
    }
	
	/**
	 * @return key iterator returning keys sorted by associated values
	 */
	public final Iterator<KeyType> sortedKeyIterator() {
		return new KeyIterator(sortedValueKeyPairs.iterator());
	}
	
	/**
	 * @return iterator over values (sorted)
	 */
	public Iterator<ValType> valueIterator() {
		return new ValueIterator(sortedValueKeyPairs.iterator());
	}

	/**
	 * get first entry in ordered list (smallest value)
	 * @return first entry
	 */
	public final Entry<KeyType,ValType> getFirst() {
		return sortedValueKeyPairs.first();
	}
	
	/**
	 * get last entry in ordered list (largest value)
	 * @return last entry
	 */
	public final Entry<KeyType,ValType> getLast() {
		return sortedValueKeyPairs.last();
	}
	
	/**
	 * remove first entry from ordered list
	 * @return first entry
	 */
	public final Entry<KeyType,ValType> removeFirst() {
		Entry<KeyType,ValType> e = sortedValueKeyPairs.first();
		sortedValueKeyPairs.remove(e);
		keyAssociations.remove(e.getKey());
		return e; 
	}
	
	/**
	 * remove last entry from ordered list
	 * @return last entry
	 */
	public final Entry<KeyType,ValType> removeLast() {
		Entry<KeyType,ValType> e = sortedValueKeyPairs.last();
		sortedValueKeyPairs.remove(e);
		keyAssociations.remove(e.getKey());
		return e; 
	}
	
	/**
     * remove entry
     * @param key key
     */
    public void remove(KeyType key) {
        ValType oldVal = keyAssociations.remove(key);
        sortedValueKeyPairs.remove(new Entry<KeyType,ValType>(key, oldVal));
    }
	
	/**
	 * return number of mapping entries
	 * @return number of entries
	 */
	public final int size() {
		return sortedValueKeyPairs.size();
	}
	
	/**
	 * Helper class used for single entries
	 * @param <KeyType> type of keys
	 * @param <ValType> type of values
	 */
	public final static class Entry<KeyType,ValType extends Comparable<ValType>>
		implements Comparable<Entry<KeyType,ValType>>
	{
		KeyType key;
        ValType value;
        private transient int hashcode = -1;
        
        /**
         * Constructor
         * @param key key
         * @param value value
         */
        public Entry(KeyType key, ValType value) {
                this.key = key;
                this.value = value;
        }
        
        /**
         * @return key
         */
        public final KeyType getKey() {
        	return key;
        }
        
        /**
         * @return value 
         */
        public final ValType getValue() {
        	return value;
        }
        
        /**
         * compare entry to an other entry
         * @param other other entry
         * @return value indicating the sort order
         */
        @Override
		@SuppressWarnings("unchecked")
		public final int compareTo(Entry<KeyType,ValType> other) {
        	if (null == other)
                return 1;	// place at end of ordering
            // compare values
        	int valueCompareResult; 
            if (other.value != null && value != null)
                valueCompareResult = value.compareTo(other.value);
            else if (other.value != null && value == null)
                valueCompareResult = -1;
            else if (other.value == null && value != null)
                valueCompareResult = 1;
            else
                valueCompareResult = 0;
            if (valueCompareResult != 0)
                return valueCompareResult;
        	else if (key instanceof Comparable)
        		// if the values are equal we must not return 0 unless
        		// the keys, too, are equal, or else new entries in the Map
        		// will overwrite entries with identical values (despite
        		// having different keys)
        		return ((Comparable<KeyType>) key).compareTo(other.key);
        	else
        		return 0;
        }
        
        /**
         * check for equality
         * @param other
         * @return whether two objects are equal
         */
        @Override
		@SuppressWarnings("unchecked")
		public final boolean equals(Object other) {
        	//return compareTo(other) == 0;
        	if (other == null)
        		return false;
        	if (!(other instanceof Entry))
        		return false;
        	//Entry<KeyType,ValType> other
        	Entry<KeyType,ValType> otherEntry = (Entry<KeyType,ValType>) other;
        	if (!value.equals(otherEntry.value))
        		return false;
        	else // if values are identical, check if keys are, too
        		return key.equals(otherEntry.key); 
        }

		/* (non-Javadoc)
		 * @see java.lang.Object#hashCode()
		 */
		@Override
		public final int hashCode() {
			if (hashcode == -1) {
				hashcode =            31 + (value != null ? value.hashCode() : 0);
				hashcode = hashcode * 31 + (key   != null ? key.hashCode()   : 0);
			}
			return hashcode;
		}
	}
	
	/**
	 * key iterator for entry pairs
	 */
	public final class KeyIterator implements Iterator<KeyType> {
		Iterator<Entry<KeyType,ValType>> baseIterator;
		
		/**
		 * @param baseIterator
		 */
		public KeyIterator(Iterator<Entry<KeyType, ValType>> baseIterator) {
			super();
			this.baseIterator = baseIterator;
		}

		@Override
		public final boolean hasNext() {
			return baseIterator.hasNext();
		}

		@Override
		public final KeyType next() {
			return baseIterator.next().key;
		}

		@Override
		public final void remove() {
			//baseIterator.remove();
			throw new UnsupportedOperationException();
		}
	}
	
	/**
	 * value iterator for entry pairs
	 */
	public final class ValueIterator implements Iterator<ValType> {
		Iterator<Entry<KeyType,ValType>> baseIterator;
		
		/**
		 * @param baseIterator
		 */
		public ValueIterator(Iterator<Entry<KeyType, ValType>> baseIterator) {
			super();
			this.baseIterator = baseIterator;
		}

		@Override
		public final boolean hasNext() {
			return baseIterator.hasNext();
		}

		@Override
		public final ValType next() {
			return baseIterator.next().value;
		}

		@Override
		public final void remove() {
			//baseIterator.remove();
			throw new UnsupportedOperationException();
		}
	}
	
	/*
	public static class ValueComparator implements java.util.Comparator {
        private Map m; // the original map
        ValueComparator() {}
        public void setMap(Map m) {this.m=m;}
        public int compare(Object o1, Object o2) {
        	// handle some exceptions here
            Comparable v1 = (Comparable) m.get(o1);
            Comparable v2 = (Comparable) m.get(o2);
            // make sure the values implement Comparable
            int result = v1.compareTo(v2);
            if (result == 0)
            	return ((Comparable) o1).compareTo(o2);
            else
            	return result;
        }
        // do something similar in equals.
	}
	*/
	
	/**
	 * main method for testing
	 * @param args arguments
	 */
	public static void main(String[] args) {
		//ValueComparator cmp = new ValueComparator();
		//TreeMap<String,Integer> t = new TreeMap<String,Integer>(cmp);
		//cmp.setMap(t);
		MapSortedByValue<String,Integer> t = new MapSortedByValue<String,Integer>(); 
		t.put("Test", 5);
		t.put("zzz", 1);
		t.put("Alpha", 9);
		t.put("Djaoid", 5);
		t.put("Beta", 5);
		t.put("Gamma", 5);
		while (t.size() != 0) {
			Entry<String,Integer> e = t.removeLast();
			System.out.println(e.getKey() + ": " + e.getValue());
		}
		/*
		for (Iterator<MapSortedByValue.Entry<String,Integer>> i = t.sortedEntrySet(); i.hasNext(); ) {
			Entry<String,Integer> e = i.next();
			System.out.println(e.getKey()+ ": " + e.getValue());
		} */
	}

}
