package intensity;


import gnu.trove.THashMap;
import gnu.trove.THashSet;
import gnu.trove.TIntDoubleHashMap;

import ilog.concert.IloException;
import ilog.concert.IloLinearNumExpr;
import ilog.concert.IloNumVar;
import ilog.concert.IloNumVarType;
import ilog.cplex.IloCplex;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.StringWriter;
import java.io.Writer;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;
import java.util.Map.Entry;
import java.util.Set;

/**
 * Linear Program 
 */
public class OptimizationProblem {
	/**
	 * optimization direction  
	 */
	public enum OptimizationDirection {
		/** minimize objective function */	MINIMIZE, 
		/** maximize objective function */  MAXIMIZE 
	}
	
	OptimizationDirection optimizationDirection;
	List<Coefficient> objectiveFunction = new LinkedList<Coefficient>();
	List<Bound> variableBounds = new LinkedList<Bound>();
	List<LinearConstraint> linearConstraints = new LinkedList<LinearConstraint>();
	int nVariables = -1;
	THashMap<Integer,String> variableNames = null;
	Set<Integer> integerConstraints = null;
	Set<Integer> binaryConstraints = null;
	
	/**
	 * Coefficient
	 */
	static public class Coefficient {
		int idx;
		double val;

		/**
		 * Constructor
		 * @param idx
		 * @param val
		 */
		public Coefficient(int idx, double val) {
			this.idx = idx;
			this.val = val;
		}
	}
	
	/**
	 * Bound
	 */
	static public class Bound {
		int idx;
		double lowerBound;
		double upperBound;

		/**
		 * Constructor
		 * @param idx
		 * @param lb lower bound
		 * @param ub upper bound
		 */
		public Bound(int idx, double lb, double ub) {
			this.idx = idx;
			this.lowerBound = lb;
			this.upperBound = ub;
		}
	}
	
	/**
	 * Constructor
	 * @param optimizationDirection optimization direction (minimization or maximization, cf. constants)
	 */
	public OptimizationProblem(OptimizationDirection optimizationDirection) {
		setOptimizationDirection(optimizationDirection);
	}
	
	
	/**
	 * @return the optimizationDirection
	 */
	public final OptimizationDirection getOptimizationDirection() {
		return optimizationDirection;
	}

	/**
	 * @param optimizationDirection the optimizationDirection to set
	 */
	public final void setOptimizationDirection(OptimizationDirection optimizationDirection) {
		this.optimizationDirection = optimizationDirection;
	}

	/**
	 * add objective function coefficient
	 * 
	 * Note: must not add a coefficient for the same variable twice!
	 * Coefficients are assumed to be zero by default
	 * 
	 * @param idx  variable index (starting with 0)
	 * @param coeff  coefficient
	 */
	public void setObjectiveCoefficient(int idx, double coeff) {
		objectiveFunction.add(new Coefficient(idx, coeff));
	}
	
	/**
	 * @return the objectiveFunction
	 */
	public final List<Coefficient> getObjectiveFunction() {
		return objectiveFunction;
	}
	
	/**
	 * add variable bounds
	 * 
	 * Note: must not add a coefficient for the same variable twice!
	 * 
	 * @param idx  variable index (starting with 0)
	 * @param min  minimum or Double.NEGATIVE_INFINITY for none
	 * @param max  maximum or Double.POSITIVE_INFINITY for none
	 */
	public void setVariableBounds(int idx, double min, double max) {
		variableBounds.add(new Bound(idx, min, max));
	}
	
	/**
	 * @return the objectiveFunction
	 */
	public final List<Bound> getVariableBounds() {
		return variableBounds;
	}
	
	/**
	 * set variable name
	 * @param idx variable index
	 * @param name variable name
	 */
	public void setVariableName(int idx, String name) {
		if (variableNames == null)
			variableNames = new THashMap<Integer,String>();
		variableNames.put(idx, name);
	}
	
	/**
	 * get variable name
	 * @param idx variable index
	 * @return variable name or null if none
	 */
	public String getVariableName(int idx) {
		if (variableNames == null)
			return null;
		return variableNames.get(idx);
	}
	
	/**
	 * get entries
	 * @return variable name entries
	 */
	public Set<Entry<Integer, String>> getVariableNames() {
		if (variableNames == null)
			return Collections.emptySet();
		return variableNames.entrySet();
	}
	
	
	/**
	 * add linear constraint
	 * @param constraint constraint
	 */
	public void addLinearConstraint(LinearConstraint constraint) {
		linearConstraints.add(constraint);
	}

	/**
	 * @return the linear constraints
	 */
	public final List<LinearConstraint> getLinearConstraints() {
		return Collections.unmodifiableList(linearConstraints);
	}
	
	/**
     * add integer/binary constraint
     * @param idx
     * @param binary whether the variable should be binary (0/1), or just any integer
     */
    public void addIntegerConstraint(int variableIdx, boolean binary) {
        if (!binary) {
            if (integerConstraints == null)
                integerConstraints = new THashSet<Integer>();
            integerConstraints.add(variableIdx);
        } else {
            if (binaryConstraints == null)
                binaryConstraints = new THashSet<Integer>();
            binaryConstraints.add(variableIdx);
        }
    }

    /**
     * @return the integer constraints
     */
    public final Set<Integer> getIntegerConstraints() {
        return Collections.unmodifiableSet(integerConstraints);
    }

    /**
     * @return the binary integer constraints
     */
    public final Set<Integer> getIntegerConstraintsBinary() {
        return Collections.unmodifiableSet(binaryConstraints);
    }

	/**
	 * Get the number of variables used (more precisely, the max. variable number plus 1).
	 * Note that this will still return the correct number if no value has been initialized explicitly.  
	 * @return the max. variable number used + 1 
	 */
	public final int getNVariables() {
		// check if a value has been given explicitly
		if (nVariables != -1)
			return nVariables;
		// otherwise determine the value on-the-fly
		int maxIdx = 0;
		for (Coefficient coeff: objectiveFunction)
			if (maxIdx < coeff.idx)
				maxIdx = coeff.idx;
		for (Bound bound: variableBounds)
			if (maxIdx < bound.idx)
				maxIdx = bound.idx;
		for (LinearConstraint constraint: linearConstraints)
			for (Coefficient coeff: constraint.coefficients)
				if (maxIdx < coeff.idx)
					maxIdx = coeff.idx;
		return maxIdx + 1;
	}

	/**
	 * Set the number of variables (or more precisely the max. variable number that will be used)
	 * 
	 * Setting this is optional as getNVariables() can also determine the number on-the-fly (which can
	 * be slightly slower though).
	 *  
	 * @param variables the max. variable number + 1 
	 */
	public final void setNVariables(int variables) {
		nVariables = variables;
	}

	
	/**
	 * Linear Program LinearConstraint
	 */
	public static class LinearConstraint {
		/**
		 * inequality direction
		 */
		public static enum Direction {
			/** less/equal **/ 		LEQ,
			/** greater/equal **/ 	GEQ,
			/** equal **/           EQ;
			
			@Override
            public String toString() {
			    switch(this) {
    			    case LEQ: return "<=";
    			    case GEQ: return ">=";
    			    case EQ:  return "=";
    			    default:  return super.toString();
			    }
			}
		}
		
		List<Coefficient> coefficients = new LinkedList<Coefficient>();
		Direction direction;
		double bound;
		String name = null;
		
		/**
		 * add objective function coefficient
		 * 
		 * Note: must not add a coefficient for the same variable twice!
		 * 
		 * @param idx  variable index (starting with 0)
		 * @param coeff  coefficient
		 */
		public void addCoefficient(int idx, double coeff) {
			coefficients.add(new Coefficient(idx, coeff));
		}
		
		/**
		 * @return the objectiveFunction
		 */
		public final List<Coefficient> getCoefficients() {
			return coefficients;
		}
		
		/**
		 * @return the direction
		 */
		public final Direction getDirection() {
			return direction;
		}

		/**
		 * @param direction the direction to set
		 */
		public final void setDirection(Direction direction) {
			this.direction = direction;
		}

		/**
		 * @return the weight
		 */
		public final double getBound() {
			return bound;
		}

		/**
		 * @param weight the weight to set
		 */
		public final void setBound(double weight) {
			this.bound = weight;
		}
		
		/**
		 * set bound (incl. direction)
		 * @param direction the direction to set
		 * @param weight the weight to set
		 */
		public final void setBound(Direction direction, double weight) {
			this.direction = direction;
			this.bound = weight;
		}

		/**
		 * @return the name
		 */
		public final String getName() {
			return name;
		}

		/**
		 * @param name the name to set
		 */
		public final void setName(String name) {
			this.name = name;
		}
		
		@Override
		public String toString() {
			return toString(null);
		}

		/**
		 * return string representation
		 * @param lp linear program for variable names or null
		 * @return string representation
		 */
		public String toString(OptimizationProblem lp) {
			StringBuilder sb = new StringBuilder();
			for (Coefficient coeff: coefficients) {
				int idx = coeff.idx;
				String var = lp.getVariableName(idx);
				if (var == null)
					var = "x" + idx;
				double coeffVal = coeff.val; 
				if (sb.length() > 0) {
					if (coeffVal >= 0)
						sb.append(" + ");
					else
						sb.append(" ");
				}
				if (coeffVal != 1.0)
					sb.append(Double.toString(coeffVal) + "*");
				sb.append(var);
			}
			sb.append(' ');
			sb.append(direction.toString());
			sb.append(' ');
			sb.append(Double.toString(bound));
			return sb.toString();
		}
				
	}
	
	/**
     * write LP format file
     * cf. http://plato.asu.edu/ftp/lp_format.txt
     * http://lpsolve.sourceforge.net/5.5/lp-format.htm
     * @param outputFile output file
     * @param cplexMode  whether to use the CPLEX format, cf. 
     *   http://lpsolve.sourceforge.net/5.5/CPLEX-format.htm
     *   instead of the lpsolve format
     * @param compact  whether to prefer a more compact notation
     * @throws IOException 
     */
    public void writeLpFile(File outputFile, boolean cplexMode, boolean compact) throws IOException {
        Writer fo = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(outputFile), "ISO-8859-1"));
        writeLpFile(fo, cplexMode, compact);
        fo.close();
    }
	
	/**
	 * write LP format file
	 * cf. http://plato.asu.edu/ftp/lp_format.txt
	 * http://lpsolve.sourceforge.net/5.5/lp-format.htm
	 * @param fo  writer object (will not be closed)
	 * @param cplexMode  whether to use the CPLEX format, cf. 
	 *   http://lpsolve.sourceforge.net/5.5/CPLEX-format.htm
	 *   instead of the lpsolve format
	 * @param compact  whether to prefer a more compact notation
	 * @throws IOException 
	 */
	public void writeLpFile(Writer fo, boolean cplexMode, boolean compact) throws IOException {
		if (cplexMode)
			fo.write(getOptimizationDirection() == OptimizationDirection.MINIMIZE ? "Min " : "Max ");
		else
			fo.write(getOptimizationDirection() == OptimizationDirection.MINIMIZE ? "min: " : "max: ");
		boolean first = true;
		for (Coefficient entry: getObjectiveFunction()) {
		    if (entry.val == 0.0)
		        continue;
			if (!first) {
				if (compact)
					fo.write((entry.val >= 0) ? "+" : "");
				else
					fo.write((entry.val >= 0) ? " + " : " ");
			} else {
				first = false;
			}
			fo.write(formatCoeff(entry.val, compact) + "x" + entry.idx);
		}
		fo.write(cplexMode ? "\nSubject to\n" : ";\n");
		
		for (LinearConstraint constraint: linearConstraints) {
			first = true;
			for (Coefficient entry: constraint.getCoefficients()) {
				if (!first){
					if (compact)
						fo.write((entry.val >= 0) ? "+" : "");
					else
						fo.write((entry.val >= 0) ? " + " : " ");
				} else {
					first = false;
				}
				fo.write(formatCoeff(entry.val, compact) + "x" + entry.idx);
			}
			String direction;
			switch (constraint.getDirection()) {
			    case GEQ: direction = " >= "; break;
			    case LEQ: direction = " <= "; break;
			    case EQ: direction = " = "; break;
			    default: throw new UnsupportedOperationException(constraint.getDirection() + " linearConstraints not supported");
			}
			fo.write(direction + formatNum(constraint.getBound()) + "\n");
		}
		
		if (cplexMode)
			fo.write("Bounds\n");
		for (Bound entry: variableBounds) {
			int idx = entry.idx;
			double lowerBound = entry.lowerBound;
			double upperBound = entry.upperBound;
			
			if (cplexMode) {
				if (lowerBound != 0) {
					if (upperBound == Double.POSITIVE_INFINITY) {
						String lb = (lowerBound == Double.NEGATIVE_INFINITY) ? "-inf" : formatNum(lowerBound);
						fo.write("x" + idx + " >= " + lb + "\n");
					} else { // upperBound != Double.POSITIVE_INFINITY
						String lb = (lowerBound == Double.NEGATIVE_INFINITY) ? "-inf" : formatNum(lowerBound);
						String ub = formatNum(upperBound);
						fo.write(lb + " <= x" + idx + " <= " + ub + "\n");
					}
				} else if (upperBound != Double.POSITIVE_INFINITY) {
					String ub = formatNum(upperBound);
					fo.write("x" + idx + " <= " + ub + "\n");
				}
			} else {
				if (lowerBound != Double.NEGATIVE_INFINITY) {
					if (upperBound == Double.POSITIVE_INFINITY)
						fo.write("x" + idx + " >= " + lowerBound + "\n");
					else // upperBound != Double.POSITIVE_INFINITY
						fo.write(lowerBound + " <= x" + idx + " <= " + upperBound + "\n");
				} else if (upperBound != Double.POSITIVE_INFINITY) {
					fo.write("x" + idx + " <= " + upperBound + "\n");
				}
			}
		}
		if (this.integerConstraints != null) {
    	    if (cplexMode) {
    	        fo.write("General\n");
    	        for (int i: this.integerConstraints)
    	            fo.write("x" + i + "\n");
    	    } else {
    	        // can also be put into a single line, separated by commas
    	        for (int i: this.integerConstraints)
                    fo.write("int x" + i + ";\n");
    	    }
		}
	    if (this.binaryConstraints != null) {
	        if (cplexMode) { 
                fo.write("Binary\n");
                for (int i: this.binaryConstraints)
                    fo.write("x" + i + "\n");
            } else {
                for (int i: this.binaryConstraints)
                    fo.write("bin x" + i + ";\n");
            }
	    }
		if (cplexMode)
			fo.write("End\n");
	}
	
	private String formatNum(double val) {
		String num = Double.toString(val);
		if (num.endsWith(".0"))
			return num;
		return num;
	}
	
	private String formatCoeff(double val, boolean compact) {
		if (val == 1)
			return "";
		if (val == -1)
			return compact ? "-" : "- ";
		String coeff = Double.toString(val);
		if (coeff.endsWith(".0")) {
			if (!compact)
				return coeff.substring(0, coeff.length() - 2) + " ";
			else
				// more compact notation for simple coefficients
				return coeff.substring(0, coeff.length() - 2);
		}
		return coeff + " ";
	}

	/**
	 * get human-readable representation
	 * @return string
	 */
	public String printableForm() {
		StringWriter w = new StringWriter();
		try {
			writeLpFile(w, true, true);
			w.close();
		} catch (IOException e) {
			return null;
		}
		return w.toString();
	}
	
	
	
	/**
	 * solve using CPLEX
	 * @param lp linear program
	 * @return  solution as a map from variables to values (note: 0 as default is 
	 *   not necessarily stored explicitly, unconstrained variables are stored as
	 *   Double.NaN)
	 * @throws IOException 
	 * @throws InterruptedException 
	 * @throws IloException 
	 */
	public static TIntDoubleHashMap solveUsingCPLEXInternal(OptimizationProblem lp) throws IOException, InterruptedException, IloException {
        IloCplex cplex = new IloCplex();
        
        // initialize variables incl. bounds
        int nVariables = lp.getNVariables();
		IloNumVar[] vars = new IloNumVar[nVariables];
		Set<Integer> integerConstraints = lp.integerConstraints;
		if (integerConstraints == null)
		    integerConstraints = Collections.emptySet();
		Set<Integer> binaryConstraints = lp.binaryConstraints;
		if (binaryConstraints == null)
		    binaryConstraints = Collections.emptySet();
		for (Bound bound: lp.getVariableBounds()) {
		    if (binaryConstraints.contains(bound.idx))
		        vars[bound.idx] = cplex.numVar(bound.lowerBound, bound.upperBound, IloNumVarType.Bool);
		    else if (integerConstraints.contains(bound.idx))
                vars[bound.idx] = cplex.intVar((int) bound.lowerBound, (int) bound.upperBound);
		    else
		        vars[bound.idx] = cplex.numVar(bound.lowerBound, bound.upperBound);
		}
		for (int i = 0; i < nVariables; i++)
			if (vars[i] == null) {
			    if (binaryConstraints.contains(i))
	                vars[i] = cplex.boolVar();
	            else if (integerConstraints.contains(i))
	                vars[i] = cplex.numVar(Double.NEGATIVE_INFINITY, Double.POSITIVE_INFINITY, IloNumVarType.Int);
	            else
	                vars[i] = cplex.numVar(Double.NEGATIVE_INFINITY, Double.POSITIVE_INFINITY);
			}
        
		// objective
		IloLinearNumExpr objective = cplex.linearNumExpr();
		for (Coefficient entry: lp.getObjectiveFunction())
			objective.addTerm(entry.val, vars[entry.idx]);
        if (lp.getOptimizationDirection() == OptimizationDirection.MINIMIZE)
        	cplex.addMinimize(objective);
        else
        	cplex.addMaximize(objective);
        
        // constraints
        for (LinearConstraint constraint: lp.getLinearConstraints()) {
        	IloLinearNumExpr linExpr = cplex.linearNumExpr();
			for (Coefficient coeff: constraint.getCoefficients())
				linExpr.addTerm(coeff.val, vars[coeff.idx]);
			switch (constraint.getDirection()) {
			    case EQ:
	                cplex.addEq(linExpr, constraint.getBound());
	                break;
			    case GEQ:
	                cplex.addGe(linExpr, constraint.getBound());
	                break;
			    case LEQ:
	                cplex.addLe(linExpr, constraint.getBound());
	                break;
	            default:
	                throw new UnsupportedOperationException(constraint.getDirection() + " linearConstraints not supported");
			}
		}
        
        
        // solve
        if (cplex.solve()) {
        	// get solutions
        	TIntDoubleHashMap solution = new TIntDoubleHashMap();
        	for (int i = 0; i < nVariables; i++)
        		try {
        			solution.put(i, cplex.getValue(vars[i]));
        		} catch (IloCplex.UnknownObjectException e) {
        			solution.put(i, Double.NaN);
        		}
            cplex.end();
            return solution;
        } else {
        	cplex.end();
        	return null;
        }
	}
}
