package intensity;

import gnu.trove.TIntDoubleHashMap;
import gnu.trove.TObjectDoubleHashMap;
import gnu.trove.TObjectIntHashMap;
import intensity.OptimizationProblem.OptimizationDirection;

import java.util.ArrayList;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import org.apache.commons.math3.util.Pair;



public class RankerILP extends Ranker {
    boolean linkSynonyms;
    
    public RankerILP(boolean linkSynonyms) {
        super();
        this.linkSynonyms = linkSynonyms;
    }

    @Override
    public Ranking rank(WordCluster cluster, Map<String, Double> wordPairWeights, TObjectDoubleHashMap<String> wordFreqs, List<List<String>> synonymSets) throws Exception 
    {
        final double RANGE = 1.0;
        final double DMIN = 0.0;
        final double C = 1.0 + RANGE * 10;
        final double synonymCoefficient = linkSynonyms ? 10000000000L : 0.000001;
    
        // determine how to re-scale scores
        double scalingFactor = 1.0;
        scalingFactor = Double.POSITIVE_INFINITY;
        for (double val: wordPairWeights.values()) {
            val = Math.abs(val);
            if (val > 0 && val < scalingFactor)
                scalingFactor = val;
        }
        if (scalingFactor == Double.POSITIVE_INFINITY)
            scalingFactor = 1.0;
        else 
            scalingFactor = 1.0/scalingFactor;
        
        // start creating LP
        OptimizationProblem lp = new OptimizationProblem(OptimizationDirection.MAXIMIZE);
        
        // assign indices to adjectives
        TObjectIntHashMap<String> adjPosIdx = new TObjectIntHashMap<String>();
        for (String adj: cluster) {
            int idx = adjPosIdx.size();
            adjPosIdx.put(adj, idx);
            lp.setVariableBounds(idx, 0.0, RANGE);
            lp.setObjectiveCoefficient(idx, 0.0);
        }
        
        if (synonymSets != null)
            for (List<String> synset: synonymSets) {
                Logger.out.print("Synset: ");
                for (String word: synset) {
                    Logger.out.print(word + " ");
                }
                Logger.out.println();
            }
        
        // set up constraints and coefficients
        int idxOfs = adjPosIdx.size();
        TObjectIntHashMap<String> deltaIdx = new TObjectIntHashMap<String>();
        for (String adj1: cluster) {
            int adj1Idx = adjPosIdx.get(adj1);
            for (String adj2: cluster)
                if (adj1.compareTo(adj2) < 0) {
                    // for each (unique) pair of adjectives, we have one delta
                    int dijIdx = idxOfs++;
                    String adjPair = adj1 + '\t' + adj2;
                    deltaIdx.put(adjPair, dijIdx);
                    double adjPairWeight = wordPairWeights.containsKey(adjPair) ? wordPairWeights.get(adjPair) * scalingFactor : 0.0;
                    Logger.out.println("x" + dijIdx + ": " + adjPair + "\t(" + adjPairWeight + ")");
                    lp.setVariableBounds(dijIdx, -RANGE, RANGE);

                    // enforce d_ij = x_j - x_i
                    int adj2Idx = adjPosIdx.get(adj2);
                    OptimizationProblem.LinearConstraint lpConstraint = new OptimizationProblem.LinearConstraint();
                    lpConstraint.addCoefficient(dijIdx, 1.0);
                    lpConstraint.addCoefficient(adj2Idx, -1.0);
                    lpConstraint.addCoefficient(adj1Idx, 1.0);
                    lpConstraint.setDirection(OptimizationProblem.LinearConstraint.Direction.EQ);
                    lpConstraint.setBound(0.0);
                    lp.addLinearConstraint(lpConstraint);
                    
                    int zidx = idxOfs++;
                    int zidx2 = idxOfs++;
                    lp.addIntegerConstraint(zidx, true);
                    lp.addIntegerConstraint(zidx2, true);
                    
                    double epsilon = 0.0001;
                    lpConstraint = new OptimizationProblem.LinearConstraint();
                    lpConstraint.addCoefficient(dijIdx, 1.0);
                    lpConstraint.addCoefficient(zidx, C);
                    lpConstraint.setDirection(OptimizationProblem.LinearConstraint.Direction.GEQ);
                    lpConstraint.setBound(DMIN);
                    lp.addLinearConstraint(lpConstraint);
                    
                    lpConstraint = new OptimizationProblem.LinearConstraint();
                    lpConstraint.addCoefficient(dijIdx, 1.0);
                    lpConstraint.addCoefficient(zidx, C);
                    lpConstraint.setDirection(OptimizationProblem.LinearConstraint.Direction.LEQ);
                    lpConstraint.setBound(DMIN + C - epsilon);
                    lp.addLinearConstraint(lpConstraint);
                    
                    lpConstraint = new OptimizationProblem.LinearConstraint();
                    lpConstraint.addCoefficient(dijIdx, 1.0);
                    lpConstraint.addCoefficient(zidx2, -C);
                    lpConstraint.setDirection(OptimizationProblem.LinearConstraint.Direction.LEQ); 
                    if (DMIN > 0.0)
                        lpConstraint.setBound(-DMIN);
                    else
                        lpConstraint.setBound(0.0);
                    lp.addLinearConstraint(lpConstraint);
                    
                    lpConstraint = new OptimizationProblem.LinearConstraint();
                    lpConstraint.addCoefficient(dijIdx, 1.0);
                    lpConstraint.addCoefficient(zidx2, -C);
                    lpConstraint.setDirection(OptimizationProblem.LinearConstraint.Direction.GEQ);
                    lpConstraint.setBound(-DMIN-C+epsilon);
                    lp.addLinearConstraint(lpConstraint);
    
                    boolean synonymPair;
                    if (!linkSynonyms)
                        synonymPair = false;
                    else 
                        synonymPair = isSynonym(synonymSets, adj1, adj2);
    
                    // set objective coefficients
                    if (linkSynonyms && synonymPair) {
                        lp.setObjectiveCoefficient(zidx,  -synonymCoefficient);
                        lp.setObjectiveCoefficient(zidx2, -synonymCoefficient);
                    } else {
                        lp.setObjectiveCoefficient(zidx, -adjPairWeight); 
                        lp.setObjectiveCoefficient(zidx2, adjPairWeight);
                    }
                }
        }
        Logger.out.println(lp.getLinearConstraints().size() + " rows (main linearConstraints)");
        Logger.out.println(lp.printableForm());
        TIntDoubleHashMap lpSolution = OptimizationProblem.solveUsingCPLEXInternal(lp);
        if (lpSolution != null) {
            Logger.out.println("LP solved.");
            
            // output ranking
            List<Pair<String,Double>> results = new ArrayList<Pair<String,Double>>(); 
            for (String adj: cluster)
                results.add(new Pair<String,Double>("x" + adjPosIdx.get(adj) + " " + adj, lpSolution.get(adjPosIdx.get(adj))));
            Collections.sort(results, new CollectionUtil.PairValueComparator<String, Double>());
            for (Pair<String,Double> adjEntry: results)
                Logger.out.println(adjEntry.getKey() + ": " + adjEntry.getValue());
            
            // extract ranking
            Ranking outputRanking = new Ranking();
            MultiValueMap<Double,String> scores = new MultiValueMap<Double,String>();
            List<Double> scoreValues = new LinkedList<Double>();
            for (String adj: cluster) {
                Double score = lpSolution.get(adjPosIdx.get(adj));
                scores.add(score, adj);
                scoreValues.add(score);
            }
            Collections.sort(scoreValues);
            double lastScore = Double.NaN;
            for (Double score: scoreValues) {
                if (score != lastScore)
                    outputRanking.add(scores.get(score), score);
                lastScore = score;
            }
            
            // additional variable output for consistency checking
            for (int variableIdx = 0; variableIdx < idxOfs; variableIdx++)
                Logger.out.println("x" + variableIdx + ": " + formatNumberDefault(lpSolution.get(variableIdx)));
    
            return outputRanking;
        } else {
            Logger.out.println("LP not solved!");
            Ranking outputRanking = new Ranking();
            outputRanking.add(new ArrayList<String>(cluster.getWords()), 0.0);
            return outputRanking;
        }
    }
}
