package intensity;
import gnu.trove.THashMap;
import gnu.trove.THashSet;
import gnu.trove.TObjectDoubleHashMap;
import gnu.trove.TObjectIntHashMap;

import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import org.apache.commons.math3.stat.correlation.SpearmansCorrelation;
import org.apache.commons.math3.util.Pair;

public class Ranking extends LinkedList<Pair<List<String>,Double>> {
    
    private static final long serialVersionUID = -3245926344204140099L;

    /**
     * Note that entries must be added in order and without duplicate adds for the
     * same score. 
     * @param words
     * @param score
     */
    public void add(Collection<String> words, Double score) {
        List<String> wordList;
        if (words instanceof List)
            wordList = (List<String>) words;
        else
            wordList = new ArrayList<String>(words);
        this.add(new Pair<List<String>, Double>(wordList, score));
    }
    
    public static final int WEAKER = 1;
    public static final int STRONGER = 2;
    public static final int EQUAL_INTENSITY = 0;
    
    public THashMap<String, Integer> getPairs() {
        THashMap<String,Integer> wordPairClassifications = new THashMap<String,Integer>();
        TObjectDoubleHashMap<String> map = getScoreMap();
        for (Object adj1x: map.keys())
            for (Object adj2x: map.keys()) {
                String adj1 = (String) adj1x;
                String adj2 = (String) adj2x;
                if (adj1.compareTo(adj2) < 0) {
                    double score1 = map.get(adj1);
                    double score2 = map.get(adj2);
                    int value = (score1 < score2) ? WEAKER : ((score1 > score2) ? STRONGER : EQUAL_INTENSITY);
                    wordPairClassifications.put(adj1 + '\t' + adj2, value);
                }
            }
        return wordPairClassifications;
    }
    
    public Set<String> getWords() {
        Set<String> set = new THashSet<String>();
        for (Pair<List<String>, Double> row: this)
            for (String adj: row.getKey())
                set.add(adj);
        return set;
    }

    public Set<String> intersectWordList(List<String> otherSet) {
        Set<String> intersection = new THashSet<String>();
        for (Pair<List<String>, Double> row: this)
            for (String adj: row.getKey())
                //System.out.println("ADJ: ;" + adj + ';');
                if (otherSet.contains(adj))
                    intersection.add(adj);
        return intersection;
    }
    
    public TObjectDoubleHashMap<String> getScoreMap() {
        TObjectDoubleHashMap<String> map = new TObjectDoubleHashMap<String>();
        for (Pair<List<String>, Double> row: this)
            for (String adj: row.getKey()) {
                if (map.containsKey(adj))
                    throw new IllegalArgumentException("Duplicate item " + adj + " in ranking!");
                map.put(adj, row.getValue());
            }
        return map;
    }
    
    public TObjectIntHashMap<String> getRankMap() {
        TObjectIntHashMap<String> map = new TObjectIntHashMap<String>();
        int rank = 1;
        for (Pair<List<String>, Double> row: this) {
            for (String adj: row.getKey()) {
                if (map.containsKey(adj))
                    throw new IllegalArgumentException("Duplicate item " + adj + " in ranking!");
                map.put(adj, rank);
            }
            rank++;
        }
        return map;
    }
    
    public double computeSpearmansCorrelationCoefficient(Ranking groundTruthRanking) {
        TObjectDoubleHashMap<String> trueScores = groundTruthRanking.getScoreMap();
        TObjectDoubleHashMap<String> ourScores = this.getScoreMap();
        double[] trueScoreArray = new double[trueScores.size()];
        double[] ourScoreArray = new double[ourScores.size()];
        int i = 0;
        for (Object adjKey: trueScores.keys()) {
            String adj = (String) adjKey;
            trueScoreArray[i] = trueScores.get(adj);
            ourScoreArray[i] = ourScores.get(adj);
            i++;
        }
        // compute Spearman Correlation using Apache Commons Math, which properly handles ties
        SpearmansCorrelation sp = new SpearmansCorrelation();
        return sp.correlation(trueScoreArray, ourScoreArray);
    }
    
    /**
     * Kendall tau correlation coefficient, implementing the variant mentioned in
     * http://research.microsoft.com/en-us/people/zhichdou/ct_ranking_paper.pdf
     * http://tripod.shef.ac.uk/publications/sanderson-soboroff%20sigir%2007.pdf
     * that accounts for ties.
     * http://en.wikipedia.org/wiki/Kendall_tau_rank_correlation_coefficient
     * @param groundTruthRanking
     * @return
     */
    public double computeKendallTauCorrelationCoefficient(Ranking groundTruthRanking) {
        TObjectDoubleHashMap<String> ourMap = getScoreMap();
        THashMap<String, Integer> groundTruthPairs = groundTruthRanking.getPairs();
        int nConcordances = 0;
        int nDiscordant = 0;
        for (java.util.Map.Entry<String, Integer> entry: groundTruthPairs.entrySet()) {
            String adjPair = entry.getKey();
            double trueClassification = entry.getValue();
            String adjs[] = adjPair.split("\t");
            double score1 = ourMap.get(adjs[0]);
            double score2 = ourMap.get(adjs[1]);
            if ((trueClassification == Ranking.WEAKER && score1 < score2) || (trueClassification == Ranking.STRONGER && score1 > score2))
                nConcordances++;
            else if ((trueClassification == Ranking.WEAKER && score1 > score2) || (trueClassification == Ranking.STRONGER && score1 < score2))
                nDiscordant++;
        }
        int groundTruthTies = groundTruthRanking.countTies();
        int ourTies = countTies();
        System.out.println(getScoreMap().size());
        System.out.println(groundTruthPairs.size());
        System.out.println(nConcordances);
        System.out.println(nDiscordant);
        System.out.println(ourTies);
        System.out.println(groundTruthTies);
        double tau = (nConcordances - nDiscordant)/Math.sqrt((nConcordances + nDiscordant + ourTies) * (nConcordances + nDiscordant + groundTruthTies));
        return tau;
    }
    
    protected int countTies() {
        int nTies = 0;
        for (Pair<List<String>, Double> row: this) {
            int nItems = row.getKey().size();
            nTies += (nItems * (nItems-1))/2;
        }
        return nTies;
    }

    /**
     * Spearman's Footrule distance
     * Cf. http://chjs.deuv.cl/Vol2N1/ChJS-02-01-01.pdf
     * @param groundTruthRanking
     * @return
     */
    public double computeSpearmanFootruleDistance(Ranking groundTruthRanking) {
        TObjectIntHashMap<String> map = getRankMap();
        TObjectIntHashMap<String> trueMap = groundTruthRanking.getRankMap();
        double d = 0;
        for (Pair<List<String>, Double> row: groundTruthRanking)
            for (String item: row.getKey()) {
                if (!map.containsKey(item)) {
                    System.err.println("Map does not contain " + item);
                    return Double.NaN;
                }
                d += Math.abs(map.get(item) - trueMap.get(item));
            }
        return d;
    }
    
    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder();
        for (Pair<List<String>, Double> row: this) {
            sb.append(row.getValue());
            sb.append("\t");
            for (String word: row.getKey()) {
                sb.append(word);
                sb.append(' ');
            }
            sb.setCharAt(sb.length() - 1, '\n'); // replace last char with newline
        }
        return sb.toString();
    }
}
