package intensity;

import gnu.trove.THashMap;
import gnu.trove.THashSet;
import gnu.trove.TObjectDoubleHashMap;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.UnsupportedEncodingException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.commons.math3.util.Pair;



public class RankingExperiments {
    private static final boolean REMOVE_INVALID_WORDS = true;
    static boolean REMOVE_ZERO_WORDS = true;
    static int SIZE_TWO_CLUSTERS = 0; // 0: ignore, 1: include, 2: only size two 
    
    /**
     * @param args
     * @throws Exception
     */
    public static void main(String[] args) throws Exception {
        File wordClusterFile = new File("data/wordnet-antonym-clusters.tsv");
        File wordFreqFile = new File("data/word-freq.txt");
        File patternOccFile = new File("data/pattern-occ-freq.txt");
        File patternFreqFile = new File("data/pattern-freq.txt");
        //File evalFile = new File("data/devset.txt");
        File evalFile = new File("data/testset.txt");
        
        LexicalStatistics stats = new LexicalStatistics(wordFreqFile, patternFreqFile); 
        
        // read evaluation file
        Map<String,List<Ranking>> evalPairsSet = readEvalFile(evalFile, false, true);
        
        for (Ranker ranker: new Ranker[] {new RankerILP(false), new RankerILP(true), new RankerDivideAndConquer(), new RankerSheinmanTokunaga()}) {
            Logger.out.println("");
            Logger.out.println("================================");
            Logger.out.println("Ranker: " + ranker.getClass().getName());
            
            evaluate(ranker, wordClusterFile, patternOccFile, stats, evalPairsSet);
        }
    }
    

    private static void evaluate(Ranker ranker, File wordClusterFile, File patternOccFile,
            LexicalStatistics stats, Map<String, List<Ranking>> evalPairsSet)
            throws UnsupportedEncodingException, FileNotFoundException,
            IOException, Exception 
   {
        Map<String, Double> wordPairWeights = stats.computeWordPairWeights(patternOccFile);
        TObjectDoubleHashMap<String> wordFreqs = stats.getWordFreqs();
        
        ClassificationEvaluation eval = new ClassificationEvaluation(null, 3);
        ClassificationEvaluation baseLineEval = new ClassificationEvaluation(null, 3);
        
        BufferedReader fi = new BufferedReader(new InputStreamReader(new FileInputStream(wordClusterFile), "UTF-8"));
        CountMap<Integer> histogram = new CountMap<Integer>();
        CountMap<Integer> testSetHistogram = new CountMap<Integer>();
        Set<Ranking> seenTestSetRankings = new THashSet<Ranking>();
        CountMap<Integer> perPartHistogram = new CountMap<Integer>();
        List<Double> kendallTaus = new LinkedList<Double>();
        List<Double> spearmanCCs = new LinkedList<Double>();
        String line;
        while ((line = fi.readLine()) != null) {
            //Logger.out.println(line);
            if (line.length() == 0)
                continue;
            // parse entry: the tab character separates the two sides of the antonymy pair
            int i = line.indexOf('\t');
            String antonymId = getAntonymId(line, i); // create an ID by using the headwords
            List<WordCluster> clusters = new ArrayList<WordCluster>();
            List<List<String>> origCluster1 = WordCluster.parseCluster(line.substring(0, i), wordFreqs);
            List<WordCluster> cluster1Partitions = WordCluster.partitionCluster(origCluster1, wordPairWeights, false);
            List<List<String>> origCluster2 = WordCluster.parseCluster(line.substring(i + 1), wordFreqs);
            List<WordCluster> cluster2Partitions = WordCluster.partitionCluster(origCluster2, wordPairWeights, false);
            perPartHistogram.increment(cluster1Partitions.size());
            perPartHistogram.increment(cluster2Partitions.size());
            clusters.addAll(cluster1Partitions);
            clusters.addAll(cluster2Partitions);
            
            List<Ranking> relevantEvalPairs = evalPairsSet.get(antonymId);
            for (WordCluster cluster: clusters) {
                // administrative issues
                Logger.out.println("--------------------------");
                Logger.out.println("Cluster size: " + cluster.size());
                histogram.increment(cluster.size());
                for (int j = 0; j<cluster.size(); j++)
                    Logger.out.println(cluster.get(j) + " -- " + wordFreqs.get(cluster.get(j)));
                
                // process cluster
                if (cluster.size() == 0)
                    Logger.out.println("--> In: " + line);
                else if (cluster.size() > 1) { // doesn't make sense to run the algorithm for singletons
                    // evaluation
                    boolean found = false;
                    if (relevantEvalPairs != null)
                        for (Ranking groundTruthRanking: relevantEvalPairs) {
                            // check if the ground truth applies to this particular cluster
                            Set<String> groundTruthWords = groundTruthRanking.getWords();
                            Set<String> coveredWords = groundTruthRanking.intersectWordList(cluster.getWords());
                            if (groundTruthWords.size() > 0 && coveredWords.size() == groundTruthWords.size()) {
                                if ((groundTruthWords.size() <= 2 && SIZE_TWO_CLUSTERS == 0) || (groundTruthWords.size() > 2 && SIZE_TWO_CLUSTERS == 2))
                                    continue;
                                if (!seenTestSetRankings.add(groundTruthRanking))
                                    continue; // do not re-evaluate existing ones
                                String id = antonymId;
                                if (cluster1Partitions.contains(cluster))
                                    id = "*** " + id;
                                if (cluster2Partitions.contains(cluster))
                                    id = id + " ***";
                                System.out.println("=== " + id.replace("\t", ", "));
                                testSetHistogram.increment(groundTruthWords.size());
                                List<List<String>> synsets = null;
                                synsets = new LinkedList<List<String>>();
                                for (Pair<List<String>, Double> row: groundTruthRanking)
                                    if (row.getKey().size() > 1)
                                        synsets.add(row.getKey());
                                WordCluster clusterToUse = cluster;
                                
                                if (REMOVE_INVALID_WORDS) {
                                    clusterToUse = new WordCluster(cluster);
                                    clusterToUse.words = new ArrayList<String>(cluster.words.size());
                                    for (String word: cluster.words)
                                        if (groundTruthWords.contains(word))
                                            clusterToUse.words.add(word);
                                }
                                
                                Ranking ranking = ranker.rank(clusterToUse, wordPairWeights, wordFreqs, synsets);
                                Logger.out.println(ranking.toString());
                                Logger.out.println("Ground Truth Ranking:\n" + groundTruthRanking);
                                TObjectDoubleHashMap<String> rankingScores = ranking.getScoreMap();
                                ClassificationEvaluation localEval = new ClassificationEvaluation(null, 3);
                                THashMap<String, Integer> pairs = groundTruthRanking.getPairs();
                                for (String wordPair: pairs.keySet()) {
                                    String words[] = wordPair.split("\t");
                                    double score1 = rankingScores.get(words[0]);
                                    double score2 = rankingScores.get(words[1]);
                                    int trueClassification = pairs.get(wordPair);
                                    int predictedClassification = score1 < score2 ? Ranking.WEAKER : (score1 > score2 ? Ranking.STRONGER : Ranking.EQUAL_INTENSITY);
                                    localEval.updateEvaluation(trueClassification, predictedClassification);
                                    eval.updateEvaluation(trueClassification, predictedClassification);
                                    double baseLineScore = wordPairWeights.containsKey(wordPair) ? wordPairWeights.get(wordPair) : (wordPairWeights.containsKey(words[1] + '\t' + words[0]) ? wordPairWeights.get(words[1] + '\t' + words[0]) : 0.0);
                                    int baseLineClassification = baseLineScore > 0 ? Ranking.WEAKER : (baseLineScore < 0 ? Ranking.STRONGER : Ranking.EQUAL_INTENSITY);
                                    baseLineEval.updateEvaluation(trueClassification, baseLineClassification);
                                    if (trueClassification != predictedClassification)
                                        Logger.out.println("Predicted " + predictedClassification + " instead of " + trueClassification + " for " + wordPair);
                                    else 
                                        Logger.out.println("Correctly predicted " + predictedClassification + " = " + trueClassification + " for " + wordPair);
                                    //if (trueClassification != baseLineClassification)
                                    //    Logger.out.println("Baseline predicted " + baseLineClassification + " instead of " + trueClassification + " for " + wordPair);
                                }
                                localEval.printResults(Logger.out);
                                
                                double tauCC = ranking.computeKendallTauCorrelationCoefficient(groundTruthRanking);
                                Logger.out.println("Kendall Tau Correlation Coefficient: " + tauCC);
                                kendallTaus.add(tauCC);
                                double spearmanCC = ranking.computeSpearmansCorrelationCoefficient(groundTruthRanking);
                                Logger.out.println("Spearman's Correlation Coefficient: " + spearmanCC);
                                spearmanCCs.add(spearmanCC);
                                Logger.out.println("Spearman's Footrule Distance: " + ranking.computeSpearmanFootruleDistance(groundTruthRanking));
                                found = true;
                            } else if (coveredWords.size() > 1 && coveredWords.size() < groundTruthWords.size()) {
                                Logger.out.println("Warning: Ground truth cluster with " + coveredWords.iterator().next() + " overlaps but is not identical");
                            }
                        }
                    if (!found) {
                        Ranking ranking = ranker.rank(cluster, wordPairWeights, wordFreqs, null);
                        Logger.out.println(ranking.toString());
                    }
                }
            }
        }
        fi.close();
        
        Logger.out.println("Results for: " + ranker.getClass().getName());
        Logger.out.begin("Baseline:");
        baseLineEval.printResults(Logger.out);
        baseLineEval.printConfusionMatrix(Logger.out);
        Logger.out.println("0: tie, 1: weaker, 2: stronger");
        Logger.out.end();
        Logger.out.begin(ranker.getClass().getName());
        eval.printResults(Logger.out);
        eval.printConfusionMatrix(Logger.out);
        Logger.out.println("Total no. of pairs:         " + eval.getNTotal());
        Logger.out.println("Total correct pairs:        " + eval.getNAgreements());
        if (kendallTaus.size() > 0) {
            Logger.out.println("Average Kendall tau:        " + sum(kendallTaus)/kendallTaus.size());
            Logger.out.println("Average Abs. |Kendall tau|: " + sumAbs(kendallTaus)/kendallTaus.size());
        }
        if (spearmanCCs.size() > 0) {
            Logger.out.println("Average Spearman CC:        " + sum(spearmanCCs)/spearmanCCs.size());
            Logger.out.println("Average Abs. |Spearman CC:| " + sumAbs(spearmanCCs)/spearmanCCs.size());
        }
        Logger.out.end();
        Logger.out.println("Partitions per Antonym Cluster Histogram");
        CountMap.printSortedByKey(perPartHistogram, Logger.out);
        Logger.out.println("Cluster Size Histogram");
        CountMap.printSortedByKey(histogram, Logger.out);
        Logger.out.println("Test Set Cluster Size Histogram");
        CountMap.printSortedByKey(testSetHistogram, Logger.out);
    }
    
    public static double sum(Collection<Double> collection) {
        double sum = 0; 
        for (double d: collection)
            sum += d;
        return sum;
    }
    
    public static double sumAbs(Collection<Double> collection) {
        double sum = 0;
        for (double d: collection)
            sum += Math.abs(d);
        return sum;
    }

    private static Map<String, List<Ranking>> readEvalFile(File evalFile, boolean ignoreSingleRankLists, boolean discardStars) throws IOException {
        BufferedReader fi = new BufferedReader(new InputStreamReader(new FileInputStream(evalFile), "UTF-8"));
        THashMap<String, List<Ranking>> evalPairsSet = new THashMap<String,List<Ranking>>();
        Ranking cEntries = new Ranking();
        String cWordPair = null;
        String line;
        while ((line = fi.readLine()) != null) {
            line = line.trim();
            if (line.startsWith("%") || line.equals(""))
                // "%" is for comments
                continue;
            else if (line.startsWith("===")) {
                // finish up last entry
                if (cEntries.size() == 1) {
                    String errorMsg = "Only one entry for: " + cWordPair + " in " + evalFile;
                    if (ignoreSingleRankLists)
                        Logger.out.println(errorMsg);
                    else {
                        fi.close();
                        throw new IOException(errorMsg);
                    }
                }
                if (cEntries.size() > 0) {
                    if (!evalPairsSet.containsKey(cWordPair))
                       evalPairsSet.put(cWordPair, new ArrayList<Ranking>());
                    evalPairsSet.get(cWordPair).add(cEntries);
                }
                // optionally start with new entry
                if (!line.startsWith("=== END")) {
                    String id = line.substring(3);
                    if (discardStars)
                        id = id.replace("**", "");
                    String[] words = id.split(",");
                    cWordPair = words[0].trim() + '\t' + words[1].trim();
                    cEntries = new Ranking();
                }
            } else {
                List<String> words = new LinkedList<String>();
                for (String word: line.split(","))
                    words.add(word.trim());
                cEntries.add(new Pair<List<String>,Double>(words, (double) cEntries.size()));
            }
        }
        fi.close();
        return evalPairsSet;
    }
    
    private static String getAntonymId(String line, int tabPos) {
        line = line.replace(';', ',');
        String word1 = line.substring(0, tabPos);
        int i = word1.indexOf(',');
        if (i != -1)
            word1 = word1.substring(0, i).trim();
        String word2 = line.substring(tabPos + 1);
        i = word2.indexOf(',');
        if (i != -1)
            word2 = word2.substring(0, i).trim();
        return word1.trim() + '\t' + word2.trim(); 
    }

}
