package intensity;
import gnu.trove.THashSet;
import gnu.trove.TObjectDoubleHashMap;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.regex.Pattern;

public class WordCluster implements Iterable<String> {
    List<String> words;
    List<List<String>> origSynsets;
    
    /**
     * @param words
     * @param origSynsets original synset, the first synset assumed to be the headword synset
     */
    public WordCluster(List<String> words, List<List<String>> origSynsets) {
        this.words = words;
        this.origSynsets = origSynsets;
    }
    
    public WordCluster() {
        // do nothing
    }
    
    public WordCluster(WordCluster clusterToCopy) {
        this.words = clusterToCopy.words;
        this.origSynsets = clusterToCopy.origSynsets;
    }

    public List<String> getWords() {
        return words;
    }

    public List<List<String>> getOrigSynsets() {
        return origSynsets;
    }
    
    @Override
    public Iterator<String> iterator() {
        return words.iterator();
    }
    
    public String get(int index) {
        return words.get(index);
    }

    public int size() {
        return words.size();
    }

    /**
     * partition cluster into connected components with respect to non-zero word pair weights
     * @param synsets list of synsets, the first synset assumed to be the headword
     * @param wordPairWeights
     * @param linkSynonymPartitions
     * @return
     */
    public static List<WordCluster> partitionCluster(List<List<String>> synsets, Map<String,Double> wordPairWeights, boolean linkSynonymPartitions) {
        List<WordCluster> clusterParts = new LinkedList<WordCluster>();
        THashSet<String> seen = new THashSet<String>();
        for (List<String> synset: synsets)
            for (String word: synset)
                if (!seen.contains(word))
                    clusterParts.add(new WordCluster(addRecursively(word, synsets, new LinkedList<String>(), seen, wordPairWeights, linkSynonymPartitions), synsets));
        return clusterParts;
    }
    
    private static LinkedList<String> addRecursively(String word,
            List<List<String>> synsets, LinkedList<String> cluster, THashSet<String> seen,
            Map<String, Double> wordPairWeights, boolean linkSynonymPartitions)
    {
        cluster.add(word);
        seen.add(word);
        for (List<String> synset: synsets)
            for (String possibleNeighbor: synset)
                if (!seen.contains(possibleNeighbor) && !word.equals(possibleNeighbor))
                    if (mapGetWithDefault(wordPairWeights, word + '\t' + possibleNeighbor, 0.0) != 0.0 || mapGetWithDefault(wordPairWeights, possibleNeighbor + '\t' + word, 0.0) != 0.0
                        || (linkSynonymPartitions && synset.contains(word)))
                          // && nodeSet.contains(possibleNeighbor), i.e. if both are in the same synset
                        addRecursively(possibleNeighbor, synsets, cluster, seen, wordPairWeights, linkSynonymPartitions);
        return cluster;
    }
    
    public static List<List<String>> parseCluster(String wordList, TObjectDoubleHashMap<String> wordFreqs) {
        // we have a number of synsets separated by ";", and each synset is a comma-separated list of words
        String[] synsetWordLists = wordList.split(";");
        List<List<String>> synsets = new ArrayList<List<String>>(synsetWordLists.length);
        for (String synsetWordList: synsetWordLists) {
            String[] candidates = synsetWordList.split(",");
            List<String> words = new ArrayList<String>(candidates.length);
            for (String candidate: candidates) {
                candidate = candidate.trim();
                if (keepWord(candidate, wordFreqs))
                    words.add(candidate);
            }
            synsets.add(words);
        }
        return synsets;
    }
    
    protected static boolean keepWord(String word, TObjectDoubleHashMap<String> wordFreqs) {
        return (!RankingExperiments.REMOVE_ZERO_WORDS || (wordFreqs.contains(word) && wordFreqs.get(word) > 0)) && !numberPattern.matcher(word).matches() && !romanNumberPattern.matcher(word).matches();
    }
    
    private static Pattern numberPattern = Pattern.compile("\\d+");
    
    private static Pattern romanNumberPattern = Pattern.compile("(i|ii|v|x|xxvii|xxxi|xlvii|l|lv|lviilx|lxv|lxxiii|c||clv|clx|d)");
        // this pattern a few of the most common Roman numbers, but definitely not all of them
    
    
    public static <K,V> V mapGetWithDefault(Map<K, V> map, K key, V defaultValue) {
        V v = map.get(key);
        if (v == null)
            return defaultValue;
        return v;
    }

    public boolean hasCycles(Map<String, Double> wordPairWeights, Logger logger) {
        boolean foundCycles = false;
        Singleton<Integer> counter = new Singleton<Integer>(0);
        for (String word: words) {
            if (hasCycles(word, Collections.singletonList(word), wordPairWeights, logger, counter)) {
                foundCycles = true;
                if (counter.getObject() > 200)
                    return true;
            }
        }
        return foundCycles;
    }

    private boolean hasCycles(String word, List<String> path, Map<String, Double> wordPairWeights, Logger logger, Singleton<Integer> counter) 
    {
        boolean foundCycles = false;
        for (String word2: words)
            if (WordCluster.mapGetWithDefault(wordPairWeights, word + '\t' + word2, 0.0) > 0.0 || WordCluster.mapGetWithDefault(wordPairWeights, word2 + '\t' + word, 0.0) < 0.0) {
                if (path.contains(word2)) {
                    if (logger != null) {
                        logger.print("Cycle in path: ");
                        for (String w: path)
                            logger.print(w + " ");
                        logger.println(word2);
                    }
                    foundCycles = true;
                    counter.setObject(counter.getObject() + 1);
                    if (counter.getObject() > 200)
                        return true;
                } else {
                    List<String> extendedPath = new ArrayList<String>(path);
                    extendedPath.add(word2);
                    if (hasCycles(word2, extendedPath, wordPairWeights, logger, counter)) {
                        foundCycles = true;
                        if (counter.getObject() > 200)
                            return true;
                    }
                }
            }
        return foundCycles;
    }
    
}

